#include <windows.h>
#include <stdio.h>
#include "nrf24spi.h"
#include "hexfile.h"

//
// nRF24XX SPI Commands
//
#define WREN       0x06 // NA             0 Set flash write enable, FSR.WEN
#define WRDIS      0x04 // NA             0 Reset flash write enable, FSR.WEN
#define RDSR       0x05 // NA             1 (or more) Read Flash Status Register(FSR)
#define WRSR       0x01 // NA             1 Write Flash Status Register(FSR).Only bits 5 and 3 (WEN and INFEN) are writable by this command.
#define READ       0x03 // Start address, 2 bytes 1 - 32768 Read data from flash
#define PROGRAM    0x02 // Start address, 2 bytes 1 - 256 Write data to flash
#define ERASE_PAGE 0x52 // Page number  , 1 byte 0 Erase addressed flash page
#define ERASE_ALL  0x62 // NA             0 Erase all pages of flash MainBlock
#define RDFPCR     0x89 // NA             1 Read Flash Protect Configuration Register (FPCR)
#define RDISIP     0x84 // NA             0 Set flash InfoPage read - back disable, FSR.RDISIP, and write 0x00 to InfoPage byte 0x22.
#define RDISMB     0x85 // NA             0 Set flash MainBlock read - back disable, FSR.RDISMB and write 0x00 to InfoPage byte 0x23.
#define ENDEBUG    0x86 // NA             0 Write 0x00 to InfoPage byte 0x24, and after next reset, the HW debugger will be enabled, see chapter 23 on page 175 for details, and FSR.DBG will be set.

//
// /write_bytes/read_bytes options
//
#define OPT_CSN            (SPI_TRANSFER_OPTIONS_CHIPSELECT_ENABLE | SPI_TRANSFER_OPTIONS_CHIPSELECT_DISABLE)
#define OPT_BYTE_INTERVALS 0x80000000

//
// write_flush options
//
#define OPT_NOADDR 0x80

static uint8  gpio_reset_pin   = 0;
static uint8  gpio_prog_pin = 0;
static uint16 gpio_directions  = 0;
static uint16 gpio_out_values  = 0;

static bool erase_infopage_ready = FALSE;

static uint8 nrf24dsys[] =
{
  0x02, 0x00, 0x06, 0x02, 0x01, 0x52, 0x75, 0x81, 0x30, 0x12, 0x02, 0x0E, 0xE5, 0x82, 0x60, 0x03,
  0x02, 0x00, 0x03, 0x79, 0x00, 0xE9, 0x44, 0x00, 0x60, 0x1B, 0x7A, 0x00, 0x90, 0x02, 0x12, 0x78,
  0x01, 0x75, 0xA0, 0x00, 0xE4, 0x93, 0xF2, 0xA3, 0x08, 0xB8, 0x00, 0x02, 0x05, 0xA0, 0xD9, 0xF4,
  0xDA, 0xF2, 0x75, 0xA0, 0xFF, 0xE4, 0x78, 0xFF, 0xF6, 0xD8, 0xFD, 0x78, 0x00, 0xE8, 0x44, 0x00,
  0x60, 0x0A, 0x79, 0x01, 0x75, 0xA0, 0x00, 0xE4, 0xF3, 0x09, 0xD8, 0xFC, 0x78, 0x00, 0xE8, 0x44,
  0x00, 0x60, 0x0C, 0x79, 0x00, 0x90, 0x00, 0x01, 0xE4, 0xF0, 0xA3, 0xD8, 0xFC, 0xD9, 0xFA, 0x02,
  0x00, 0x03, 0xAF, 0x82, 0xA2, 0xAF, 0x92, 0xD5, 0xC2, 0xAF, 0xAE, 0x8E, 0x75, 0x8E, 0x01, 0x75,
  0xFA, 0xAA, 0x75, 0xFA, 0x55, 0xE5, 0x28, 0x60, 0x03, 0x43, 0x87, 0x10, 0xD2, 0xFD, 0x8F, 0xFA,
  0x20, 0xFC, 0xFD, 0xC2, 0xFD, 0xE5, 0x28, 0x60, 0x03, 0x53, 0x87, 0xEF, 0x8E, 0x8E, 0xA2, 0xD5,
  0x92, 0xAF, 0x22, 0xAE, 0x82, 0xAF, 0x83, 0xA2, 0xAF, 0x92, 0xD5, 0xC2, 0xAF, 0x85, 0x8E, 0x2E,
  0x75, 0x8E, 0x01, 0x75, 0xFA, 0xAA, 0x75, 0xFA, 0x55, 0xE5, 0x2D, 0x60, 0x03, 0x43, 0x87, 0x10,
  0xD2, 0xFD, 0x8E, 0x2F, 0x8F, 0x30, 0xAA, 0x28, 0xAB, 0x29, 0xAC, 0x2A, 0xA8, 0x2B, 0xA9, 0x2C,
  0x88, 0x05, 0x89, 0x07, 0x18, 0xB8, 0xFF, 0x01, 0x19, 0xED, 0x4F, 0x60, 0x26, 0xAE, 0x2F, 0xAF,
  0x30, 0x05, 0x2F, 0xE4, 0xB5, 0x2F, 0x02, 0x05, 0x30, 0x8A, 0x82, 0x8B, 0x83, 0x8C, 0xF0, 0x12,
  0x01, 0xF2, 0xFD, 0xA3, 0xAA, 0x82, 0xAB, 0x83, 0x8E, 0x82, 0x8F, 0x83, 0xED, 0xF0, 0x30, 0xFC,
  0xCF, 0x80, 0xFB, 0xC2, 0xFD, 0xE5, 0x2D, 0x60, 0x03, 0x53, 0x87, 0xEF, 0x85, 0x2E, 0x8E, 0xA2,
  0xD5, 0x92, 0xAF, 0x22, 0xAE, 0x82, 0xAF, 0x83, 0xE5, 0x2D, 0x60, 0x02, 0xD2, 0xFB, 0x8E, 0x2E,
  0x8F, 0x2F, 0xAB, 0x28, 0xAC, 0x29, 0xAD, 0x2A, 0xA9, 0x2B, 0xAA, 0x2C, 0x89, 0x00, 0x8A, 0x07,
  0x19, 0xB9, 0xFF, 0x01, 0x1A, 0xE8, 0x4F, 0x60, 0x22, 0xAE, 0x2E, 0xAF, 0x2F, 0x05, 0x2E, 0xE4,
  0xB5, 0x2E, 0x02, 0x05, 0x2F, 0x8E, 0x82, 0x8F, 0x83, 0xE0, 0xFF, 0x8B, 0x82, 0x8C, 0x83, 0x8D,
  0xF0, 0x12, 0x01, 0xD7, 0xA3, 0xAB, 0x82, 0xAC, 0x83, 0x80, 0xD1, 0xE5, 0x2D, 0x60, 0x02, 0xC2,
  0xFB, 0x22, 0xC2, 0xAF, 0x75, 0x28, 0x08, 0x75, 0x29, 0x00, 0x75, 0x2A, 0x40, 0x75, 0x2B, 0x20,
  0x75, 0x2C, 0x00, 0x75, 0x2D, 0x01, 0x90, 0x00, 0x00, 0x12, 0x01, 0x04, 0x74, 0x70, 0xB5, 0xBC,
  0x33, 0xD2, 0xCE, 0x75, 0xA6, 0x01, 0x75, 0xBC, 0x01, 0xC2, 0xC2, 0x10, 0xC2, 0x02, 0x80, 0xFB,
  0xE5, 0xBE, 0x30, 0xE4, 0xF6, 0x85, 0x08, 0xBF, 0x7F, 0x01, 0xBF, 0x20, 0x00, 0x50, 0xEC, 0x8F,
  0x06, 0x0F, 0xEE, 0x24, 0x08, 0xF9, 0x87, 0xBF, 0x10, 0xC2, 0x02, 0x80, 0xFB, 0xE5, 0xBE, 0x20,
  0xE5, 0xD9, 0x80, 0xE6, 0x75, 0xC9, 0x02, 0xD2, 0xCE, 0x75, 0xA6, 0x01, 0x75, 0xBC, 0x01, 0xC2,
  0xC2, 0x10, 0xC2, 0x02, 0x80, 0xFB, 0xE5, 0xBE, 0x30, 0xE1, 0xF6, 0x7F, 0x00, 0xBF, 0x20, 0x00,
  0x50, 0xEF, 0x8F, 0x06, 0x0F, 0xEE, 0x24, 0x08, 0xF9, 0x87, 0xBD, 0x10, 0xC2, 0x02, 0x80, 0xFB,
  0xE5, 0xBE, 0x20, 0xE2, 0xDC, 0x80, 0xE6, 0x20, 0xF7, 0x11, 0x30, 0xF6, 0x13, 0x88, 0x83, 0xA8,
  0x82, 0x20, 0xF5, 0x09, 0xF6, 0xA8, 0x83, 0x75, 0x83, 0x00, 0x22, 0x80, 0xFE, 0xF2, 0x80, 0xF5,
  0xF0, 0x22, 0x20, 0xF7, 0x14, 0x30, 0xF6, 0x14, 0x88, 0x83, 0xA8, 0x82, 0x20, 0xF5, 0x07, 0xE6,
  0xA8, 0x83, 0x75, 0x83, 0x00, 0x22, 0xE2, 0x80, 0xF7, 0xE4, 0x93, 0x22, 0xE0, 0x22, 0x75, 0x82,
  0x00, 0x22, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF
};

FT_STATUS pin_mode(FT_HANDLE handle, uint8 pin, uint8 mode)
{
  FT_STATUS status;

  if (mode == OUTPUT)
    gpio_directions |= (1 << pin);
  else
    gpio_directions &= ~(1 << pin);
  if (!FT_SUCCESS(status = FT_WriteGPIO(handle, gpio_directions, gpio_out_values)))
    printf("FT_WriteGPIO() = %d\n", status);
  return status;
}

FT_STATUS pin_write(FT_HANDLE handle, uint8 pin, uint8 value)
{
  FT_STATUS status = FT_OK;

  if (value)
    gpio_out_values |= (1 << pin);
  else
    gpio_out_values &= ~(1 << pin);
  if (gpio_directions & (1 << pin))
  {
    if (!FT_SUCCESS(status = FT_WriteGPIO(handle, gpio_directions, gpio_out_values)))
      printf("FT_WriteGPIO() = %d\n", status);
  }
  return status;
}

FT_STATUS pin_read(FT_HANDLE handle, uint8 pin, uint8 *value)
{
  FT_STATUS status;
  uint16 val;

  if (!FT_SUCCESS(status = FT_ReadGPIO(handle, &val)))
    printf("FT_ReadGPIO() = %d\n", status);
  else
    *value = (val & (1 << pin) ? HIGH : LOW);
  return status;
}

FT_STATUS reset(FT_HANDLE handle, bool state)
{
  return pin_write(handle, gpio_reset_pin, state ? LOW : HIGH);
}

FT_STATUS prog(FT_HANDLE handle, bool state)
{
  return pin_write(handle, gpio_prog_pin, state ? HIGH : LOW);
}

FT_STATUS program_mode(FT_HANDLE handle, bool enable)
{
  FT_STATUS status;

  erase_infopage_ready = FALSE;

  status = reset(handle, TRUE);
  if (enable)
  {
    status = prog(handle, enable);
    reset(handle, FALSE);
    Sleep(2);
  }
  return status;
}

static FT_STATUS firmware_start(FT_HANDLE handle)
{
  FT_STATUS status;

  erase_infopage_ready = FALSE;

  reset(handle, TRUE);
  status = prog(handle, FALSE);
  reset(handle, FALSE);
  Sleep(100);
  return status;
}

static FT_STATUS toggle_cs(FT_HANDLE handle, bool state)
{
  FT_STATUS status;

  if (!FT_SUCCESS(status = SPI_ToggleCS(handle, state)))
    printf("SPI_ToggleCS() = %d\n", status);
  return status;
}

static FT_STATUS write_bytes(FT_HANDLE handle, uint8 *buffer, uint32 buflen, uint32 options)
{
  FT_STATUS status;
  uint32 sizeTransferred;

  if (options & SPI_TRANSFER_OPTIONS_CHIPSELECT_ENABLE)
    toggle_cs(handle, TRUE);
    
  while (buflen)
  {
    if (!FT_SUCCESS(status = SPI_Write(handle, buffer, (options & OPT_BYTE_INTERVALS ? 1 : buflen), &sizeTransferred, 0)))
    {
      printf("SPI_Write() = %d\n", status);
      break;
    }
    else if (sizeTransferred == 0)
    {
      printf("SPI_Write() IO Error\n");
      return FT_IO_ERROR;
    }
    buffer += sizeTransferred;
    buflen -= sizeTransferred;
  }

  if (options & SPI_TRANSFER_OPTIONS_CHIPSELECT_DISABLE)
    toggle_cs(handle, FALSE);
  
  return status;
}

static FT_STATUS read_bytes(FT_HANDLE handle, uint8 *buffer, uint32 buflen, uint32 options)
{
  FT_STATUS status;
  uint32 sizeTransferred;

  if (options & SPI_TRANSFER_OPTIONS_CHIPSELECT_ENABLE)
    toggle_cs(handle, TRUE);

  while (buflen)
  {
    if (!FT_SUCCESS(status = SPI_Read(handle, buffer, (options & OPT_BYTE_INTERVALS ? 1 : buflen), &sizeTransferred, 0)))
    {
      printf("SPI_Read() = %d\n", status);
      break;
    }
    else if (sizeTransferred == 0)
    {
      printf("SPI_Read() IO Error\n");
      return FT_IO_ERROR;
    }
    buffer += sizeTransferred;
    buflen -= sizeTransferred;
  }

  if (options & SPI_TRANSFER_OPTIONS_CHIPSELECT_DISABLE)
    toggle_cs(handle, FALSE);

  return status;
}

static FT_STATUS write_and_read_bytes(FT_HANDLE handle, uint8 *command, uint32 cmdlen, uint8 *buffer, uint32 buflen)
{
  FT_STATUS status;

  // enable CS
  toggle_cs(handle, TRUE);

  if (FT_SUCCESS(status = write_bytes(handle, command, cmdlen, 0)))
    status = read_bytes(handle, buffer, buflen, 0);

  // disable CS
  toggle_cs(handle, FALSE);

  return status;
}

FT_STATUS read_fsr(FT_HANDLE handle, uint8 *value)
{
  uint8 command = RDSR;
  return write_and_read_bytes(handle, &command, sizeof(command), value, sizeof(*value));
}

FT_STATUS read_fpcr(FT_HANDLE handle, uint8 *value)
{
  uint8 command = RDFPCR;
  return write_and_read_bytes(handle, &command, sizeof(command), value, sizeof(*value));
}

FT_STATUS read_flush(FT_HANDLE handle, uint16 addr, uint8 *buffer, uint16 buflen, uint8 options)
{
  FT_STATUS status;

  // INFEN
  uint8 cmd_infen[] = { WRSR, (uint8)(options & OPT_INFOPAGE ? FSR_INFEN : 0) };
  if (FT_SUCCESS(status = write_bytes(handle, cmd_infen, sizeof(cmd_infen), OPT_CSN)))
  {
    uint8 cmd_read[] = { READ, (uint8)HIBYTE(addr), (uint8)LOBYTE(addr) };
    status = write_and_read_bytes(handle, cmd_read, sizeof(cmd_read), buffer, buflen);
  }
  return status;
}

FT_STATUS write_flush(FT_HANDLE handle, uint16 addr, uint8 *buffer, uint16 buflen, uint8 options)
{
  FT_STATUS status = FT_OK;

  while (FT_SUCCESS(status) && buflen)
  {
    // INFEN and WEN
    uint8 cmd_wren[] = { WRSR, (uint8)(options & OPT_INFOPAGE ? FSR_INFEN | FSR_WEN : FSR_WEN) };
    if (FT_SUCCESS(status = write_bytes(handle, cmd_wren, sizeof(cmd_wren), OPT_CSN)))
    {
      // enable CS
      toggle_cs(handle, TRUE);
      // program address
      if (!(options & OPT_NOADDR))
      {
        uint8 cmd_program[] = { PROGRAM, (uint8)HIBYTE(addr), (uint8)LOBYTE(addr) };
        status = write_bytes(handle, cmd_program, sizeof(cmd_program), 0);
        if (!(options & OPT_INFOPAGE) && (addr < sizeof(nrf24dsys)))
          erase_infopage_ready = FALSE;
      }
      // program or erase
      if (FT_SUCCESS(status))
      {
        uint16 len = (buflen <= 256 ? buflen : 256);
        if (FT_SUCCESS(status = write_bytes(handle, buffer, len, 0)))
        {
          addr   += len;
          buffer += len;
          buflen -= len;
        }
      }
      // disable CS
      toggle_cs(handle, FALSE);
      // wait busy
      if (FT_SUCCESS(status))
      {
        uint8 value;
        while (FT_SUCCESS(read_fsr(handle, &value)))
        {
          if ((value & FSR_RDYN) == 0)
            break;
        }
      }
    }
  }
  return status;
}

FT_STATUS enable_debug(FT_HANDLE handle)
{
  uint8 cmd_endebug[] = { ENDEBUG };
  return write_flush(handle, 0, cmd_endebug, sizeof(cmd_endebug), OPT_NOADDR);
}

FT_STATUS readback_disable(FT_HANDLE handle, uint8 options)
{
  FT_STATUS status;
  uint8 cmd_rdisxx[] = { (uint8)(options & OPT_INFOPAGE ? RDISIP : RDISMB) };
  if (FT_SUCCESS(status = write_flush(handle, 0, cmd_rdisxx, sizeof(cmd_rdisxx), OPT_NOADDR)))
  {
    if (!(options & OPT_INFOPAGE))
      erase_infopage_ready = FALSE;
  }
  return status;
}

FT_STATUS erase_all(FT_HANDLE handle)
{
  FT_STATUS status;
  uint8 cmd_erase_all[] = { ERASE_ALL };
  if (FT_SUCCESS(status = write_flush(handle, 0, cmd_erase_all, sizeof(cmd_erase_all), OPT_NOADDR)))
    erase_infopage_ready = TRUE;
  return status;
}

FT_STATUS erase_page(FT_HANDLE handle, uint8 pageno, uint8 options, const char *device)
{
  FT_STATUS status;
  uint8 fsr, dsys[DSYS_SIZE];

  // backup dsys
  if (options & OPT_INFOPAGE)
  {
    // read FSR
    if (!FT_SUCCESS(status = read_fsr(handle, &fsr)))
      return status;
    // erase ok ?
    if ((fsr & (FSR_RDISMB | FSR_RDISIP)) && !erase_infopage_ready)
      return FT_OTHER_ERROR;
    // read dsys
    if (fsr & FSR_RDISIP)
    {
      // write dsys read code
      if (!FT_SUCCESS(status = write_flush(handle, 0x0000, nrf24dsys, sizeof(nrf24dsys), 0)))
        return status;
      // run code. and read dsys from SPI
      firmware_start(handle);
      read_bytes(handle, dsys, sizeof(dsys), OPT_CSN | OPT_BYTE_INTERVALS);
      program_mode(handle, TRUE);
      // erase all
      if (!FT_SUCCESS(status = erase_all(handle)))
        return status;
    }
    else
    {
      if (!FT_SUCCESS(status = read_flush(handle, 0x0000, dsys, sizeof(dsys), options)))
        return status;
    }
    // dsys backup to file
    save_dsys_to_file(device, dsys, sizeof(dsys), 16);
    // infopage is 0 page only
    pageno = 0;
  }
  // erase page
  uint8 cmd_erase_page[] = { ERASE_PAGE, pageno };
  if (FT_SUCCESS(status = write_flush(handle, 0, cmd_erase_page, sizeof(cmd_erase_page), OPT_NOADDR | options)))
  {
    if (options & OPT_INFOPAGE)
    {
      // restart
      firmware_start(handle);
      program_mode(handle, TRUE);
      // reprogram dsys
      if (!(options & OPT_DSYS))
        status = write_flush(handle, 0, dsys, sizeof(dsys), options);
    }
  }
  return status;
}

FT_STATUS get_usb_device_count(uint32 *count)
{
  FT_STATUS status;

  if (!FT_SUCCESS(status = SPI_GetNumChannels(count)))
  {
    if (status == FT_NOT_SUPPORTED)
      printf("Can not find \"libMPSSE.dll\" or \"ftd2xx.dll\" for execution.\n");
    else
      printf("SPI_GetNumChannels() = %d\n", status);
  }
  return status;
}

FT_STATUS get_usb_device_info(uint32 index, FT_DEVICE_LIST_INFO_NODE *info)
{
  FT_STATUS status;

  if (!FT_SUCCESS(status = SPI_GetChannelInfo(index, info)))
    printf("SPI_GetChannelInfo(%d) = %d\n", index, status);
  return status;
}

FT_STATUS open_usb_device(FT_HANDLE *handle, uint32 index)
{
  FT_STATUS status;

  if (!FT_SUCCESS(status = SPI_OpenChannel(index, handle)))
    printf("SPI_OpenChannel(%d) = %d\n", index, status);
  return status;
}

FT_STATUS close_usb_device(FT_HANDLE handle)
{
  FT_STATUS status;

  if (!FT_SUCCESS(status = SPI_CloseChannel(handle)))
    printf("SPI_CloseChannel() = %d\n", status);
  return status;
}

FT_STATUS init_spi(FT_HANDLE handle)
{
  FT_STATUS status;
  ChannelConfig channelConf;

  gpio_reset_pin  = SPI_RESET;
  gpio_prog_pin   = SPI_PROG;
  gpio_directions = (1 << gpio_reset_pin) | (1 << gpio_prog_pin);
  gpio_out_values = (0 << gpio_reset_pin) | (1 << gpio_prog_pin);
  //
  channelConf.ClockRate     = SPI_CLOCK_RATE;
  channelConf.LatencyTimer  = SPI_LATENCY_TIMER;
  channelConf.configOptions = SPI_CONFIG_OPTION_MODE0 | ((SPI_CSN - 3) << 2) | SPI_CONFIG_OPTION_CS_ACTIVELOW;
  channelConf.Pin = ((gpio_directions & 0xFF) << 0) | ((gpio_out_values & 0xFF) << 8);
  if (!FT_SUCCESS(status = SPI_InitChannel(handle, &channelConf)))
    printf("SPI_InitChannel() = %d\n", status);
  return status;
}
